from typing import Dict, Optional, Union

Fraction = float
Weight = float
Probability = float
Distribution = Union[
    # For example, 1 means 100% reads.
    int,
    # For example, 0.25 means 25% reads.
    float,
    # For example, {0.25: 1, 0.8: 2} means 25% reads one third of the time and
    # 80% reads two thirds of the time.
    Dict[Fraction, Weight],
]

def canonicalize(d: Distribution) -> Dict[Fraction, Probability]:
    if isinstance(d, int):
        if d < 0 or d > 1:
            raise ValueError('distribution must be in the range [0, 1]')
        return {float(d): 1.}
    elif isinstance(d, float):
        if d < 0 or d > 1:
            raise ValueError('distribution must be in the range [0, 1]')
        return {d: 1.}
    elif isinstance(d, dict):
        if len(d) == 0:
            raise ValueError('distribution cannot empty')

        if any(weight < 0 for weight in d.values()):
            raise ValueError('distribution cannot have negative weights')

        total_weight = sum(d.values())
        if total_weight == 0:
            raise ValueError('distribution cannot have zero weight')

        return {float(f): weight / total_weight
                for (f, weight) in d.items()
                if weight > 0}
    else:
        raise ValueError('distribution must be an int, a float, a Dict[float, '
                         'float] or a List[Tuple[float, float]]')


def canonicalize_rw(read_fraction: Optional[Distribution],
                    write_fraction: Optional[Distribution]) \
                    -> Dict[Fraction, Probability]:
    if read_fraction is None and write_fraction is None:
        raise ValueError('Either read_fraction or write_fraction must be given')
    elif read_fraction is not None and write_fraction is not None:
        raise ValueError('Only one of read_fraction or write_fraction can be '
                         'given')
    elif read_fraction is not None:
        return canonicalize(read_fraction)
    else:
        assert write_fraction is not None
        return {1 - f: weight
                for (f, weight) in canonicalize(write_fraction).items()}



from .expr import Node, choose, majority
from .quorum_system import NoStrategyFoundError, QuorumSystem, Strategy
from .search import NoQuorumSystemFoundError, search
from .viz import (
    plot_node_load,
    plot_node_load_on,
    plot_node_utilization,
    plot_node_utilization_on,
    plot_node_throughput,
    plot_node_throughput_on,
    plot_load_distribution,
    plot_load_distribution_on,
)



from . import distribution
from . import geometry
from .distribution import Distribution
from .expr import Node
from .geometry import Point, Segment
from .quorum_system import Strategy
from typing import Dict, FrozenSet, List, Optional, Set, Tuple, TypeVar
import collections
import matplotlib
import matplotlib.pyplot as plt


T = TypeVar('T')


def plot_node_load(filename: str,
                   strategy: Strategy[T],
                   nodes: Optional[List[Node[T]]] = None,
                   read_fraction: Optional[Distribution] = None,
                   write_fraction: Optional[Distribution] = None):
    fig, ax = plt.subplots()
    plot_node_load_on(ax, strategy, nodes, read_fraction, write_fraction)
    ax.set_xlabel('Node')
    ax.set_ylabel('Load')
    fig.tight_layout()
    fig.savefig(filename)


def plot_node_load_on(ax: plt.Axes,
                      strategy: Strategy[T],
                      nodes: Optional[List[Node[T]]] = None,
                      read_fraction: Optional[Distribution] = None,
                      write_fraction: Optional[Distribution] = None):
    _plot_node_load_on(ax,
                       strategy,
                       nodes or list(strategy.nodes()),
                       scale=1,
                       scale_by_node_capacity=True,
                       read_fraction=read_fraction,
                       write_fraction=write_fraction)


def plot_node_utilization(filename: str,
                          strategy: Strategy[T],
                          nodes: Optional[List[Node[T]]] = None,
                          read_fraction: Optional[Distribution] = None,
                          write_fraction: Optional[Distribution] = None):
    fig, ax = plt.subplots()
    plot_node_utilization_on(ax, strategy, nodes, read_fraction, write_fraction)
    ax.set_xlabel('Node')
    ax.set_ylabel('Utilization')
    fig.tight_layout()
    fig.savefig(filename)


def plot_node_utilization_on(ax: plt.Axes,
                      strategy: Strategy[T],
                      nodes: Optional[List[Node[T]]] = None,
                      read_fraction: Optional[Distribution] = None,
                      write_fraction: Optional[Distribution] = None):
    _plot_node_load_on(ax,
                       strategy,
                       nodes or list(strategy.nodes()),
                       scale=strategy.capacity(read_fraction, write_fraction),
                       scale_by_node_capacity=True,
                       read_fraction=read_fraction,
                       write_fraction=write_fraction)


def plot_node_throughput(filename: str,
                   strategy: Strategy[T],
                   nodes: Optional[List[Node[T]]] = None,
                   read_fraction: Optional[Distribution] = None,
                   write_fraction: Optional[Distribution] = None):
    fig, ax = plt.subplots()
    plot_node_throughput_on(ax, strategy, nodes, read_fraction, write_fraction)
    ax.set_xlabel('Node')
    ax.set_ylabel('Throughput')
    fig.tight_layout()
    fig.savefig(filename)


def plot_node_throughput_on(ax: plt.Axes,
                            strategy: Strategy[T],
                            nodes: Optional[List[Node[T]]] = None,
                            read_fraction: Optional[Distribution] = None,
                            write_fraction: Optional[Distribution] = None,
                            draw_node_capacities: bool = True):
    nodes = nodes or list(strategy.nodes())
    d = distribution.canonicalize_rw(read_fraction, write_fraction)
    fr = sum(weight * fr for (fr, weight) in d.items())
    fw = 1 - fr
    # TODO(mwhittaker): Explain.
    node_limits = [
        fr * strategy.x_read_probability[node.x] / node_load +
        fw * strategy.x_write_probability[node.x] / node_load
        for node in nodes
        for node_load in [strategy.node_load(
            node,
            read_fraction=read_fraction,
            write_fraction=write_fraction
        )]
    ]
    _plot_node_load_on(ax,
                       strategy,
                       nodes,
                       scale=strategy.capacity(read_fraction, write_fraction),
                       scale_by_node_capacity=False,
                       read_fraction=read_fraction,
                       write_fraction=write_fraction,
                       node_limits=node_limits if draw_node_capacities else None)


def _plot_node_load_on(ax: plt.Axes,
                       sigma: Strategy[T],
                       nodes: List[Node[T]],
                       scale: float,
                       scale_by_node_capacity: bool,
                       read_fraction: Optional[Distribution] = None,
                       write_fraction: Optional[Distribution] = None,
                       node_limits: List[float] = None):
    d = distribution.canonicalize_rw(read_fraction, write_fraction)
    x_list = [node.x for node in nodes]
    x_index = {x: i for (i, x) in enumerate(x_list)}
    x_ticks = list(range(len(x_list)))

    def one_hot(quorum: FrozenSet[T]) -> List[float]:
        bar_heights = [0.0] * len(x_list)
        for x in quorum:
            bar_heights[x_index[x]] = 1.
        return bar_heights

    width = 0.8
    def plot_quorums(sigma: Dict[FrozenSet[T], float],
                     fraction: float,
                     bottoms: List[float],
                     capacities: List[float],
                     cmap: matplotlib.colors.Colormap):
        for (i, (quorum, weight)) in enumerate(sigma.items()):
            bar_heights = [scale * fraction * weight * x for x in one_hot(quorum)]
            if scale_by_node_capacity:
                for j in range(len(bar_heights)):
                    bar_heights[j] /= capacities[j]

            ax.bar(x_ticks,
                   bar_heights,
                   bottom=bottoms,
                   color=cmap(0.75 - i * 0.5 / len(sigma)),
                   edgecolor='white', width=width)

            for j, (bar_height, bottom) in enumerate(zip(bar_heights, bottoms)):
                text = ''.join(str(x) for x in sorted(list(quorum))) # type: ignore
                if bar_height != 0:
                    ax.text(x_ticks[j], bottom + bar_height / 2, text,
                            ha='center', va='center')

            for j in range(len(bar_heights)):
                bottoms[j] += bar_heights[j]

    # Plot the quorums.
    fr = sum(weight * fr for (fr, weight) in d.items())
    fw = 1 - fr
    read_capacities = [node.read_capacity for node in nodes]
    write_capacities = [node.write_capacity for node in nodes]
    bottoms = [0.0] * len(x_list)
    plot_quorums(sigma.sigma_r, fr, bottoms, read_capacities,
                 matplotlib.cm.get_cmap('Reds'))
    plot_quorums(sigma.sigma_w, fw, bottoms,
                 write_capacities, matplotlib.cm.get_cmap('Blues'))

    # Plot the limits, if there are any.
    if node_limits is not None:
        for (i, limit) in enumerate(node_limits):
            ax.plot([i - width/2, i + width/2], [limit, limit], color='black')

    ax.set_xticks(x_ticks)
    ax.set_xticklabels(str(x) for x in x_list)


def plot_load_distribution(filename: str,
                           strategy: Strategy[T],
                           nodes: Optional[List[Node[T]]] = None):
    fig, ax = plt.subplots()
    plot_load_distribution_on(ax, strategy, nodes)
    ax.set_xlabel('Read Fraction')
    ax.set_ylabel('Load')
    fig.tight_layout()
    fig.savefig(filename)


def _group(segments: Dict[T, Segment]) -> Dict[Segment, List[T]]:
    groups: Dict[Segment, List[T]] = collections.defaultdict(list)
    for x, segment in segments.items():
        matches = (s for s in groups if segment.approximately_equal(s))
        groups[next(matches, segment)].append(x)
    return groups


def plot_load_distribution_on(ax: plt.Axes,
                              strategy: Strategy[T],
                              nodes: Optional[List[Node[T]]] = None):
    nodes = nodes or list(strategy.nodes())

    # We want to plot every node's load distribution. Multiple nodes might
    # have the same load distribution, so we group the nodes by their
    # distribution. The grouping is a little annoying because two floats
    # might not be exactly equal but pretty close.
    groups = _group({
        node.x: Segment(
            Point(0, strategy.node_load(node, read_fraction=0)),
            Point(1, strategy.node_load(node, read_fraction=1))
        )
        for node in nodes
    })

    # Compute and plot the max of all segments. We plot the load first so that
    # it lies underneath the node loads.
    path = geometry.max_of_segments(list(groups.keys()))
    ax.plot([p[0] for p in path],
            [p[1] for p in path],
            label='load',
            linewidth=4)

    # We plot the node loads second so that they appear above the load.
    for segment, xs in groups.items():
        ax.plot([segment.l.x, segment.r.x],
                [segment.l.y, segment.r.y],
                '--',
                label=','.join(str(x) for x in xs),
                linewidth=2,
                alpha=0.75)



from quoracle import *
from quoracle.geometry import *
from typing import Any, Callable, List, NamedTuple, Optional, Tuple
import unittest


class TestGeometry(unittest.TestCase):
    def test_eq(self):
        l = Point(0, 1)
        r = Point(1, 1)
        m = Point(0.5, 0.5)
        self.assertEqual(Segment(l, r), Segment(l, r))
        self.assertNotEqual(Segment(l, r), Segment(l, m))

    def test_compatible(self):
        s1 = Segment(Point(0, 1), Point(1, 2))
        s2 = Segment(Point(0, 2), Point(1, 1))
        s3 = Segment(Point(0.5, 2), Point(1, 1))
        self.assertTrue(s1.compatible(s2))
        self.assertTrue(s2.compatible(s1))
        self.assertFalse(s1.compatible(s3))
        self.assertFalse(s3.compatible(s1))
        self.assertFalse(s2.compatible(s3))
        self.assertFalse(s3.compatible(s2))

    def test_call(self):
        segment = Segment(Point(0, 0), Point(1, 1))
        for x in [0.0, 0.25, 0.5, 0.75, 1.0]:
            self.assertEqual(segment(x), x)

        segment = Segment(Point(0, 0), Point(1, 2))
        for x in [0.0, 0.25, 0.5, 0.75, 1.0]:
            self.assertEqual(segment(x), 2*x)

        segment = Segment(Point(1, 2), Point(3, 6))
        for x in [1.0, 1.25, 1.5, 1.75, 2.0, 2.25, 2.5, 2.75, 3.0]:
            self.assertEqual(segment(x), 2*x)

        segment = Segment(Point(0, 1), Point(1, 0))
        for x in [0.0, 0.25, 0.5, 0.75, 1.0]:
            self.assertEqual(segment(x), 1 - x)

    def test_slope(self):
        self.assertEqual(Segment(Point(0, 0), Point(1, 1)).slope(), 1.0)
        self.assertEqual(Segment(Point(0, 1), Point(1, 2)).slope(), 1.0)
        self.assertEqual(Segment(Point(1, 1), Point(2, 2)).slope(), 1.0)
        self.assertEqual(Segment(Point(1, 1), Point(2, 3)).slope(), 2.0)
        self.assertEqual(Segment(Point(1, 1), Point(2, 0)).slope(), -1.0)

    def test_above(self):
        s1 = Segment(Point(0, 0), Point(1, 0.5))
        s2 = Segment(Point(0, 0.5), Point(1, 2))
        s3 = Segment(Point(0, 1.5), Point(1, 0.5))

        self.assertFalse(s1.above(s1))
        self.assertFalse(s1.above(s2))
        self.assertFalse(s1.above(s3))

        self.assertTrue(s2.above(s1))
        self.assertFalse(s2.above(s2))
        self.assertFalse(s2.above(s3))

        self.assertTrue(s3.above(s1))
        self.assertFalse(s3.above(s2))
        self.assertFalse(s3.above(s3))

    def test_above_eq(self):
        s1 = Segment(Point(0, 0), Point(1, 0.5))
        s2 = Segment(Point(0, 0.5), Point(1, 2))
        s3 = Segment(Point(0, 1.5), Point(1, 0.5))

        self.assertTrue(s1.above_eq(s1))
        self.assertFalse(s1.above_eq(s2))
        self.assertFalse(s1.above_eq(s3))

        self.assertTrue(s2.above_eq(s1))
        self.assertTrue(s2.above_eq(s2))
        self.assertFalse(s2.above_eq(s3))

        self.assertTrue(s3.above_eq(s1))
        self.assertFalse(s3.above_eq(s2))
        self.assertTrue(s3.above_eq(s3))

    def test_intersects(self):
        s1 = Segment(Point(0, 0), Point(1, 0.5))
        s2 = Segment(Point(0, 0.5), Point(1, 2))
        s3 = Segment(Point(0, 1.5), Point(1, 0.5))

        self.assertTrue(s1.intersects(s1))
        self.assertFalse(s1.intersects(s2))
        self.assertTrue(s1.intersects(s3))

        self.assertFalse(s2.intersects(s1))
        self.assertTrue(s2.intersects(s2))
        self.assertTrue(s2.intersects(s3))

        self.assertTrue(s3.intersects(s1))
        self.assertTrue(s3.intersects(s2))
        self.assertTrue(s3.intersects(s3))

    def test_intersection(self):
        s1 = Segment(Point(0, 0), Point(1, 1))
        s2 = Segment(Point(0, 1), Point(1, 0))
        s3 = Segment(Point(0, 1), Point(1, 1))
        s4 = Segment(Point(0, 0.25), Point(1, 0.25))

        self.assertEqual(s1.intersection(s1), None)
        self.assertEqual(s1.intersection(s2), Point(0.5, 0.5))
        self.assertEqual(s1.intersection(s3), Point(1, 1))
        self.assertEqual(s1.intersection(s4), Point(0.25, 0.25))

        self.assertEqual(s2.intersection(s1), Point(0.5, 0.5))
        self.assertEqual(s2.intersection(s2), None)
        self.assertEqual(s2.intersection(s3), Point(0, 1))
        self.assertEqual(s2.intersection(s4), Point(0.75, 0.25))

        self.assertEqual(s3.intersection(s1), Point(1, 1))
        self.assertEqual(s3.intersection(s2), Point(0, 1))
        self.assertEqual(s3.intersection(s3), None)
        self.assertEqual(s3.intersection(s4), None)

        self.assertEqual(s4.intersection(s1), Point(0.25, 0.25))
        self.assertEqual(s4.intersection(s2), Point(0.75, 0.25))
        self.assertEqual(s4.intersection(s3), None)
        self.assertEqual(s4.intersection(s4), None)

    def test_max_one_segment(self):
        s1 = Segment(Point(0, 0), Point(1, 1))
        s2 = Segment(Point(0, 1), Point(1, 0))
        s3 = Segment(Point(0, 1), Point(1, 1))
        s4 = Segment(Point(0, 0.25), Point(1, 0.25))
        s5 = Segment(Point(0, 0.75), Point(1, 0.75))

        def is_subset(xs: List[Any], ys: List[Any]) -> bool:
            return all(x in ys for x in xs)

        for s in [s1, s2, s3, s4, s5]:
            self.assertEqual(max_of_segments([s]), [s.l, s.r])

        expected = [
            ([s1, s1], [(0, 0), (1, 1)]),
            ([s1, s2], [(0, 1), (0.5, 0.5), (1, 1)]),
            ([s1, s3], [(0, 1), (1, 1)]),
            ([s1, s4], [(0, 0.25), (0.25, 0.25), (1, 1)]),
            ([s1, s5], [(0, 0.75), (0.75, 0.75), (1, 1)]),
            ([s2, s2], [(0, 1), (1, 0)]),
            ([s2, s3], [(0, 1), (1, 1)]),
            ([s2, s4], [(0, 1), (0.75, 0.25), (1, 0.25)]),
            ([s2, s5], [(0, 1), (0.25, 0.75), (1, 0.75)]),
            ([s3, s3], [(0, 1), (1, 1)]),
            ([s3, s4], [(0, 1), (1, 1)]),
            ([s3, s5], [(0, 1), (1, 1)]),
            ([s4, s4], [(0, 0.25), (1, 0.25)]),
            ([s4, s5], [(0, 0.75), (1, 0.75)]),
            ([s5, s5], [(0, 0.75), (1, 0.75)]),

            ([s1, s2, s4], [(0, 1), (0.5, 0.5), (1, 1)]),
            ([s1, s2, s5], [(0, 1), (0.25, 0.75), (0.75, 0.75), (1, 1)]),
        ]
        for segments, path in expected:
            self.assertTrue(is_subset(path, max_of_segments(segments)))
            self.assertTrue(is_subset(path, max_of_segments(segments[::-1])))



from quoracle import *
from quoracle.quorum_system import *
import unittest


class TestStrategy(unittest.TestCase):
    def test_get_quorum(self) -> None:
        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')
        e = Node('e')

        for reads in [a,
                      a + b,
                      a + b + c,
                      choose(2, [a, b, c]),
                      choose(2, [a, b, c, d, e]),
                      (a + b) * (c + d),
                      (a * b) + (c * d)]:
            qs = QuorumSystem(reads=reads)
            sigma = qs.uniform_strategy()
            for _ in range(10):
                self.assertTrue(qs.is_read_quorum(sigma.get_read_quorum()))
                self.assertTrue(qs.is_write_quorum(sigma.get_write_quorum()))

    def test_load_cap_util(self) -> None:
        a = Node('a', write_capacity=10, read_capacity=50)
        b = Node('b', write_capacity=20, read_capacity=60)
        c = Node('c', write_capacity=30, read_capacity=70)
        d = Node('d', write_capacity=40, read_capacity=80)

        qs = QuorumSystem(reads=a*b + c*d)
        sigma = qs.make_strategy(
            sigma_r = {
                frozenset({'a', 'b'}): 0.75,
                frozenset({'c', 'd'}): 0.25,
            },
            sigma_w = {
                frozenset({'a', 'c'}): 0.1,
                frozenset({'a', 'd'}): 0.2,
                frozenset({'b', 'c'}): 0.3,
                frozenset({'b', 'd'}): 0.4,
            },
        )

        node_loads_08 = {
            a: 0.8 / 50 * 0.75 + 0.2 / 10 * (0.1 + 0.2),
            b: 0.8 / 60 * 0.75 + 0.2 / 20 * (0.3 + 0.4),
            c: 0.8 / 70 * 0.25 + 0.2 / 30 * (0.1 + 0.3),
            d: 0.8 / 80 * 0.25 + 0.2 / 40 * (0.2 + 0.4),
        }
        load_08 = max(node_loads_08.values())
        cap_08 = 1 / load_08
        node_throughputs_08 = {
            a: cap_08 * (0.8 * 0.75 + 0.2 * (0.1 + 0.2)),
            b: cap_08 * (0.8 * 0.75 + 0.2 * (0.3 + 0.4)),
            c: cap_08 * (0.8 * 0.25 + 0.2 * (0.1 + 0.3)),
            d: cap_08 * (0.8 * 0.25 + 0.2 * (0.2 + 0.4)),
        }
        self.assertAlmostEqual(sigma.load(read_fraction=0.8), load_08)
        self.assertAlmostEqual(sigma.capacity(read_fraction=0.8), cap_08)
        for node, l in node_loads_08.items():
            self.assertAlmostEqual(sigma.node_load(node, read_fraction=0.8), l)
            self.assertAlmostEqual(
                sigma.node_utilization(node, read_fraction=0.8),
                l * cap_08)
        for node, t in node_throughputs_08.items():
            self.assertAlmostEqual(
                sigma.node_throughput(node, read_fraction=0.8),
                t)

        node_loads_05 = {
            a: 0.5 / 50 * 0.75 + 0.5 / 10 * (0.1 + 0.2),
            b: 0.5 / 60 * 0.75 + 0.5 / 20 * (0.3 + 0.4),
            c: 0.5 / 70 * 0.25 + 0.5 / 30 * (0.1 + 0.3),
            d: 0.5 / 80 * 0.25 + 0.5 / 40 * (0.2 + 0.4),
        }
        load_05 = max(node_loads_05.values())
        cap_05 = 1 / load_05
        node_throughputs_05 = {
            a: cap_05 * (0.5 * 0.75 + 0.5 * (0.1 + 0.2)),
            b: cap_05 * (0.5 * 0.75 + 0.5 * (0.3 + 0.4)),
            c: cap_05 * (0.5 * 0.25 + 0.5 * (0.1 + 0.3)),
            d: cap_05 * (0.5 * 0.25 + 0.5 * (0.2 + 0.4)),
        }
        self.assertAlmostEqual(sigma.load(read_fraction=0.5), load_05)
        self.assertAlmostEqual(sigma.capacity(read_fraction=0.5), cap_05)
        for node, l in node_loads_05.items():
            self.assertAlmostEqual(sigma.node_load(node, read_fraction=0.5), l)
            self.assertAlmostEqual(
                sigma.node_utilization(node, read_fraction=0.5),
                l * cap_05)
        for node, t in node_throughputs_05.items():
            self.assertAlmostEqual(
                sigma.node_throughput(node, read_fraction=0.5),
                t)

        fr = {0.8: 0.7, 0.5: 0.3}
        node_loads = {
            a: 0.7 * (0.8 / 50 * 0.75 + 0.2 / 10 * (0.1 + 0.2)) +
               0.3 * (0.5 / 50 * 0.75 + 0.5 / 10 * (0.1 + 0.2)),
            b: 0.7 * (0.8 / 60 * 0.75 + 0.2 / 20 * (0.3 + 0.4)) +
               0.3 * (0.5 / 60 * 0.75 + 0.5 / 20 * (0.3 + 0.4)),
            c: 0.7 * (0.8 / 70 * 0.25 + 0.2 / 30 * (0.1 + 0.3)) +
               0.3 * (0.5 / 70 * 0.25 + 0.5 / 30 * (0.1 + 0.3)),
            d: 0.7 * (0.8 / 80 * 0.25 + 0.2 / 40 * (0.2 + 0.4)) +
               0.3 * (0.5 / 80 * 0.25 + 0.5 / 40 * (0.2 + 0.4)),
        }
        load = (0.7 * max(node_loads_08.values()) +
                0.3 * max(node_loads_05.values()))
        cap = (0.7 * 1 / max(node_loads_08.values()) +
                0.3 * 1/ max(node_loads_05.values()))
        self.assertAlmostEqual(sigma.load(read_fraction=fr), load)
        self.assertAlmostEqual(sigma.capacity(read_fraction=fr), cap)
        node_throughputs = {
            a: cap_08 * 0.7 * (0.8 * 0.75 + 0.2 * (0.1 + 0.2)) +
               cap_05 * 0.3 * (0.5 * 0.75 + 0.5 * (0.1 + 0.2)),
            b: cap_08 * 0.7 * (0.8 * 0.75 + 0.2 * (0.3 + 0.4)) +
               cap_05 * 0.3 * (0.5 * 0.75 + 0.5 * (0.3 + 0.4)),
            c: cap_08 * 0.7 * (0.8 * 0.25 + 0.2 * (0.1 + 0.3)) +
               cap_05 * 0.3 * (0.5 * 0.25 + 0.5 * (0.1 + 0.3)),
            d: cap_08 * 0.7 * (0.8 * 0.25 + 0.2 * (0.2 + 0.4)) +
               cap_05 * 0.3 * (0.5 * 0.25 + 0.5 * (0.2 + 0.4)),
        }
        for node, l in node_loads.items():
            self.assertAlmostEqual(sigma.node_load(node, read_fraction=fr), l)
            self.assertAlmostEqual(
                sigma.node_utilization(node, read_fraction=fr),
                0.7 * cap_08 * node_loads_08[node] +
                0.3 * cap_05 * node_loads_05[node])
        for node, t in node_throughputs.items():
            self.assertAlmostEqual(
                sigma.node_throughput(node, read_fraction=fr),
                t)

    def test_network_load(self) -> None:
        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')
        e = Node('e')

        qs = QuorumSystem(reads=a*b + c*d*e)
        sigma = qs.make_strategy(
            sigma_r = {
                frozenset({'a', 'b'}): 75,
                frozenset({'c', 'd', 'e'}): 25,
            },
            sigma_w = {
                frozenset({'a', 'c'}): 5,
                frozenset({'a', 'd'}): 10,
                frozenset({'a', 'e'}): 15,
                frozenset({'b', 'c'}): 20,
                frozenset({'b', 'd'}): 25,
                frozenset({'b', 'e'}): 25,
            },
        )

        self.assertEqual(sigma.network_load(read_fraction=0.8),
            0.8 * 0.75 * 2 +
            0.8 * 0.25 * 3 +
            0.2 * 2
        )

    def test_latency(self) -> None:
        a = Node('a', latency=datetime.timedelta(seconds=1))
        b = Node('b', latency=datetime.timedelta(seconds=2))
        c = Node('c', latency=datetime.timedelta(seconds=3))
        d = Node('d', latency=datetime.timedelta(seconds=4))
        e = Node('e', latency=datetime.timedelta(seconds=5))

        qs = QuorumSystem(reads=a*b + c*d*e)
        sigma = qs.make_strategy(
            sigma_r = {
                frozenset({'a', 'b'}): 10,
                frozenset({'a', 'b', 'c'}): 20,
                frozenset({'c', 'd', 'e'}): 30,
                frozenset({'c', 'd', 'e', 'a'}): 40,
            },
            sigma_w = {
                frozenset({'a', 'c'}): 5,
                frozenset({'a', 'd'}): 10,
                frozenset({'a', 'e'}): 15,
                frozenset({'b', 'c'}): 20,
                frozenset({'b', 'd'}): 25,
                frozenset({'b', 'e'}): 25,
            },
        )

        self.assertEqual(sigma.latency(read_fraction=0.8),
            0.8 * 0.10 * datetime.timedelta(seconds=2) +
            0.8 * 0.20 * datetime.timedelta(seconds=2) +
            0.8 * 0.30 * datetime.timedelta(seconds=5) +
            0.8 * 0.40 * datetime.timedelta(seconds=5) +
            0.2 * 0.05 * datetime.timedelta(seconds=3) +
            0.2 * 0.10 * datetime.timedelta(seconds=4) +
            0.2 * 0.15 * datetime.timedelta(seconds=5) +
            0.2 * 0.20 * datetime.timedelta(seconds=3) +
            0.2 * 0.25 * datetime.timedelta(seconds=4) +
            0.2 * 0.25 * datetime.timedelta(seconds=5)
        )



from . import distribution
from . import geometry
from .distribution import Distribution
from .expr import Expr, Node
from .geometry import Point, Segment
from typing import *
import collections
import datetime
import itertools
import math
import pulp
import random


T = TypeVar('T')

LOAD = 'load'
NETWORK = 'network'
LATENCY = 'latency'


class NoStrategyFoundError(ValueError):
    pass


class QuorumSystem(Generic[T]):
    def __init__(self, reads: Optional[Expr[T]] = None,
                       writes: Optional[Expr[T]] = None) -> None:
        if reads is not None and writes is not None:
            optimal_writes = reads.dual()
            if not all(optimal_writes.is_quorum(wq) for wq in writes.quorums()):
                raise ValueError(
                    'Not all read quorums intersect all write quorums')

            self.reads = reads
            self.writes = writes
        elif reads is not None and writes is None:
            self.reads = reads
            self.writes = reads.dual()
        elif reads is None and writes is not None:
            self.reads = writes.dual()
            self.writes = writes
        else:
            raise ValueError('A QuorumSystem must be instantiated with a set '
                             'of read quorums or a set of write quorums')

        self.x_to_node = {node.x: node for node in self.nodes()}

    def __repr__(self) -> str:
        return f'QuorumSystem(reads={self.reads}, writes={self.writes})'

    def read_quorums(self) -> Iterator[Set[T]]:
        return self.reads.quorums()

    def write_quorums(self) -> Iterator[Set[T]]:
        return self.writes.quorums()

    def is_read_quorum(self, xs: Set[T]) -> bool:
        return self.reads.is_quorum(xs)

    def is_write_quorum(self, xs: Set[T]) -> bool:
        return self.writes.is_quorum(xs)

    def node(self, x: T) -> Node[T]:
        return self.x_to_node[x]

    def nodes(self) -> Set[Node[T]]:
        return self.reads.nodes() | self.writes.nodes()

    def elements(self) -> Set[T]:
        return {node.x for node in self.nodes()}

    def resilience(self) -> int:
        return min(self.read_resilience(), self.write_resilience())

    def read_resilience(self) -> int:
        return self.reads.resilience()

    def write_resilience(self) -> int:
        return self.writes.resilience()

    def dup_free(self) -> bool:
        return self.reads.dup_free() and self.writes.dup_free()

    def load(self,
             optimize: str = LOAD,
             load_limit: Optional[float] = None,
             network_limit: Optional[float] = None,
             latency_limit: Optional[datetime.timedelta] = None,
             read_fraction: Optional[Distribution] = None,
             write_fraction: Optional[Distribution] = None,
             f: int = 0) -> float:
        return self.strategy(
            optimize=optimize,
            load_limit=load_limit,
            network_limit=network_limit,
            latency_limit=latency_limit,
            read_fraction=read_fraction,
            write_fraction=write_fraction,
            f=f
        ).load(read_fraction, write_fraction)

    def capacity(self,
                 optimize: str = LOAD,
                 load_limit: Optional[float] = None,
                 network_limit: Optional[float] = None,
                 latency_limit: Optional[datetime.timedelta] = None,
                 read_fraction: Optional[Distribution] = None,
                 write_fraction: Optional[Distribution] = None,
                 f: int = 0) -> float:
        return self.strategy(
            optimize=optimize,
            load_limit=load_limit,
            network_limit=network_limit,
            latency_limit=latency_limit,
            read_fraction=read_fraction,
            write_fraction=write_fraction,
            f=f
        ).capacity(read_fraction, write_fraction)

    def network_load(self,
                     optimize: str = LOAD,
                     load_limit: Optional[float] = None,
                     network_limit: Optional[float] = None,
                     latency_limit: Optional[datetime.timedelta] = None,
                     read_fraction: Optional[Distribution] = None,
                     write_fraction: Optional[Distribution] = None,
                     f: int = 0) -> float:
        return self.strategy(
            optimize=optimize,
            load_limit=load_limit,
            network_limit=network_limit,
            latency_limit=latency_limit,
            read_fraction=read_fraction,
            write_fraction=write_fraction,
            f=f
        ).network_load(read_fraction, write_fraction)

    def latency(self,
                optimize: str = LOAD,
                load_limit: Optional[float] = None,
                network_limit: Optional[float] = None,
                latency_limit: Optional[datetime.timedelta] = None,
                read_fraction: Optional[Distribution] = None,
                write_fraction: Optional[Distribution] = None,
                f: int = 0) -> float:
        return self.strategy(
            optimize=optimize,
            load_limit=load_limit,
            network_limit=network_limit,
            latency_limit=latency_limit,
            read_fraction=read_fraction,
            write_fraction=write_fraction,
            f=f
        ).latency(read_fraction, write_fraction)

    def uniform_strategy(self, f: int = 0) -> 'Strategy[T]':
        """
        uniform_strategy(f) returns a uniform strategy over the minimal
        f-resilient quorums. That is, every minimal f-resilient quorum is
        equally likely to be chosen.
        """
        if f < 0:
            raise ValueError('f must be >= 0')
        elif f == 0:
            read_quorums = list(self.read_quorums())
            write_quorums = list(self.write_quorums())
        else:
            xs = list(self.elements())
            read_quorums = list(self._f_resilient_quorums(f, xs, self.reads))
            write_quorums = list(self._f_resilient_quorums(f, xs, self.reads))
            if len(read_quorums) == 0:
                raise NoStrategyFoundError(
                    f'There are no {f}-resilient read quorums')
            if len(write_quorums) == 0:
                raise NoStrategyFoundError(
                    f'There are no {f}-resilient write quorums')

        read_quorums = self._minimize(read_quorums)
        write_quorums = self._minimize(write_quorums)
        sigma_r = {frozenset(q): 1 / len(read_quorums) for q in read_quorums}
        sigma_w = {frozenset(q): 1 / len(write_quorums) for q in write_quorums}
        return Strategy(self, sigma_r, sigma_w)

    def make_strategy(self,
                      sigma_r: Dict[FrozenSet[T], float],
                      sigma_w: Dict[FrozenSet[T], float]) -> 'Strategy[T]':
        if not all(0 <= weight for weight in sigma_r.values()):
            raise ValueError('sigma_r has negative weights')
        if not all(0 <= weight for weight in sigma_w.values()):
            raise ValueError('sigma_w has negative weights')
        if not all(self.is_read_quorum(set(rq)) for rq in sigma_r):
            raise ValueError('sigma_r has non-read quorums')
        if not all(self.is_write_quorum(set(wq)) for wq in sigma_w):
            raise ValueError('sigma_w has non-write quorums')
        normalized_sigma_r = {rq: weight / sum(sigma_r.values())
                              for (rq, weight) in sigma_r.items()}
        normalized_sigma_w = {wq: weight / sum(sigma_w.values())
                              for (wq, weight) in sigma_w.items()}
        return Strategy(self,
                        sigma_r=normalized_sigma_r,
                        sigma_w=normalized_sigma_w)

    def strategy(self,
                 optimize: str = LOAD,
                 load_limit: Optional[float] = None,
                 network_limit: Optional[float] = None,
                 latency_limit: Optional[datetime.timedelta] = None,
                 read_fraction: Optional[Distribution] = None,
                 write_fraction: Optional[Distribution] = None,
                 f: int = 0) -> 'Strategy[T]':
        if optimize not in {LOAD, NETWORK, LATENCY}:
            raise ValueError(
                f'optimize must be one of {LOAD}, {NETWORK}, or {LATENCY}')

        if optimize == LOAD and load_limit is not None:
            raise ValueError(
                'a load limit cannot be set when optimizing for load')

        if optimize == NETWORK and network_limit is not None:
            raise ValueError(
                'a network limit cannot be set when optimizing for network')

        if optimize == LATENCY and latency_limit is not None:
            raise ValueError(
                'a latency limit cannot be set when optimizing for latency')

        if f < 0:
            raise ValueError('f must be >= 0')

        d = distribution.canonicalize_rw(read_fraction, write_fraction)
        if f == 0:
            return self._load_optimal_strategy(
                list(self.read_quorums()),
                list(self.write_quorums()),
                d,
                optimize=optimize,
                load_limit=load_limit,
                network_limit=network_limit,
                latency_limit=latency_limit)
        else:
            xs = list(self.elements())
            read_quorums = list(self._f_resilient_quorums(f, xs, self.reads))
            write_quorums = list(self._f_resilient_quorums(f, xs, self.writes))
            if len(read_quorums) == 0:
                raise NoStrategyFoundError(
                    f'There are no {f}-resilient read quorums')
            if len(write_quorums) == 0:
                raise NoStrategyFoundError(
                    f'There are no {f}-resilient write quorums')
            return self._load_optimal_strategy(
                read_quorums,
                write_quorums,
                d,
                optimize=optimize,
                load_limit=load_limit,
                network_limit=network_limit,
                latency_limit=latency_limit)

    def _minimize(self, sets: List[Set[T]]) -> List[Set[T]]:
        sets = sorted(sets, key=lambda s: len(s))
        minimal_elements: List[Set[T]] = []
        for x in sets:
            if not any(x >= y for y in minimal_elements):
                minimal_elements.append(x)
        return minimal_elements

    def _f_resilient_quorums(self,
                             f: int,
                             xs: List[T],
                             e: Expr) -> Iterator[Set[T]]:
        """
        Consider a set X of elements in xs. We say X is f-resilient if, despite
        removing an arbitrary set of f elements from X, X is a quorum in e.
        _f_resilient_quorums returns the set of all f-resilient quorums.
        """
        assert f >= 1

        def helper(s: Set[T], i: int) -> Iterator[Set[T]]:
            if all(e.is_quorum(s - set(failure))
                   for failure in itertools.combinations(s, min(f, len(s)))):
                yield set(s)
                return

            for j in range(i, len(xs)):
                s.add(xs[j])
                yield from helper(s, j + 1)
                s.remove(xs[j])

        return helper(set(), 0)

    def _read_quorum_latency(self, quorum: Set[Node[T]]) -> datetime.timedelta:
        return self._quorum_latency(quorum, self.is_read_quorum)

    def _write_quorum_latency(self, quorum: Set[Node[T]]) -> datetime.timedelta:
        return self._quorum_latency(quorum, self.is_write_quorum)

    def _quorum_latency(self,
                        quorum: Set[Node[T]],
                        is_quorum: Callable[[Set[T]], bool]) \
                        -> datetime.timedelta:
        nodes = list(quorum)
        nodes.sort(key=lambda node: node.latency)
        for i in range(len(quorum)):
            if is_quorum({node.x for node in nodes[:i+1]}):
                return nodes[i].latency
        raise ValueError('_quorum_latency called on a non-quorum')

    def _load_optimal_strategy(
            self,
            read_quorums: List[Set[T]],
            write_quorums: List[Set[T]],
            read_fraction: Dict[float, float],
            optimize: str = LOAD,
            load_limit: Optional[float] = None,
            network_limit: Optional[float] = None,
            latency_limit: Optional[datetime.timedelta] = None) -> 'Strategy[T]':
        """
        Consider the following 2x2 grid quorum system.

            a   b

            c   d

        with

            read_quorums = [{a, b}, {c, d}]
            write_quorums = [{a, c}, {a, d}, {b, c}, {b, d}]

        We want to find the strategy that is optimal with respect to load,
        network load, or latency that satisfies the provided load, network
        load, or latency constraints.

        We can find the optimal strategy using linear programming. First, we
        create a variable ri for every read quorum i and a variable wi for
        every write quorum i. ri represents the probabilty of selecting the ith
        read quorum, and wi represents the probabilty of selecting the ith
        write quorum.

        We now explain how to represent load, network load, and latency as
        linear expressions.

        Load
        ====
        Assume a read fraction fr and write fraction fw. The load of a node a is

            load(a) = (fr * rprob(a) / rcap(a)) + (fw * wprob(a) / wcap(a))

        where prob_r(a) and prob_w(a) are the probabilities that a is selected
        as part of a read or write quorum respectively; and rcap(a) and wcap(a)
        are the read and write capacities of a. We can express prob_r(a) and
        prob_w(a) as follows:

                  rprob(a) = sum({ri | a is in read quorum i})
                  wprob(a) = sum({wi | a is in write quorum i})

        Using the example grid quorum above, we have:

                      rprob(a) = r0    wprob(a) = w0 + w1
                      rprob(b) = r0    wprob(b) = w2 + w3
                      rprob(c) = r1    wprob(c) = w0 + w2
                      rprob(d) = r1    wprob(d) = w1 + w3

        The load of a strategy is the maximum load on any node. We can compute
        this by minimizing a new variable l and constraining the load of every
        node to be less than l. Using the example above, we have

               min l subject to
               fr * r0 * rcap(a) + fw * (w0 + w1) * wcap(a) <= l
               fr * r0 * rcap(b) + fw * (w2 + w3) * wcap(b) <= l
               fr * r1 * rcap(c) + fw * (w0 + w2) * wcap(c) <= l
               fr * r1 * rcap(d) + fw * (w1 + w3) * wcap(d) <= l

        To compute the load of a strategy with respect to a distribution of
        read_fractions, we compute the load for every value of fr and weight
        according to the distribution. For example, imagine fr is 0.9 80% of
        the time and 0.5 20% of the time. We have:

               min 0.8 * l0.9 + 0.2 * l0.5
               0.9 * r0 * rcap(a) + 0.1 * (w0 + w1) * wcap(a) <= l0.9
               0.9 * r0 * rcap(b) + 0.1 * (w2 + w3) * wcap(b) <= l0.9
               0.9 * r1 * rcap(c) + 0.1 * (w0 + w2) * wcap(c) <= l0.9
               0.9 * r1 * rcap(d) + 0.1 * (w1 + w3) * wcap(d) <= l0.9
               0.5 * r0 * rcap(a) + 0.5 * (w0 + w1) * wcap(a) <= l0.5
               0.5 * r0 * rcap(b) + 0.5 * (w2 + w3) * wcap(b) <= l0.5
               0.5 * r1 * rcap(c) + 0.5 * (w0 + w2) * wcap(c) <= l0.5
               0.5 * r1 * rcap(d) + 0.5 * (w1 + w3) * wcap(d) <= l0.5

        Let the expression for load be LOAD.

        Network
        =======
        The network load of a strategy is the expected size of a quorum. For a
        fixed fr, We can compute the network load as:

                     fr * sum_i(size(read quorum i) * ri) +
                     fw * sum_i(size(write quorum i) * ri)

        Using the example above:

               fr * (2*r0 + 2*r1) + fw * (2*w0 + 2*w1 + 2*w2 + 2*w3)

        For a distribution of read fractions, we compute the weighted average.
        Let the expression for network load be NETWORK.

        Latency
        =======
        The latency of a strategy is the expected latency of a quorum. We can
        compute the latency as:

                     fr * sum_i(latency(read quorum i) * ri) +
                     fw * sum_i(latency(write quorum i) * ri)

        Using the example above (assuming every node has a latency of 1):

               fr * (1*r0 + 1*r1) + fw * (1*w0 + 1*w1 + 1*w2 + 1*w3)

        For a distribution of read fractions, we compute the weighted average.
        Let the expression for latency be LATENCY.

        Linear Program
        ==============
        To find an optimal strategy, we use a linear program. The objective
        specified by the user is minimized, and any provided constraints are
        added as constraints to the program. For example, imagine the user
        wants a load optimal strategy with network load <= 2 and latency <= 3.
        We form the program:

            min LOAD subject to
            sum_i(ri) = 1 # ensure we have a valid distribution on read quorums
            sum_i(wi) = 1 # ensure we have a valid distribution on write quorums
            NETWORK <= 2
            LATENCY <= 3

        Using the example above assuming a fixed fr, we have:

               min l subject to
               fr * r0 * rcap(a) + fw * (w0 + w1) * wcap(a) <= l
               fr * r0 * rcap(b) + fw * (w2 + w3) * wcap(b) <= l
               fr * r1 * rcap(c) + fw * (w0 + w2) * wcap(c) <= l
               fr * r1 * rcap(d) + fw * (w1 + w3) * wcap(d) <= l
               fr * (2*r0 + 2*r1) + fw * (2*w0 + 2*w1 + 2*w2 + 2*w3) <= 2
               fr * (1*r0 + 1*r1) + fw * (1*w0 + 1*w1 + 1*w2 + 1*w3) <= 3

        If we instead wanted to minimize network load with load <= 4 and
        latency <= 5, we would have the following program:

               min fr * (2*r0 + 2*r1) +
                   fw * (2*w0 + 2*w1 + 2*w2 + 2*w3) subject to
               fr * r0 * rcap(a) + fw * (w0 + w1) * wcap(a) <= 4
               fr * r0 * rcap(b) + fw * (w2 + w3) * wcap(b) <= 4
               fr * r1 * rcap(c) + fw * (w0 + w2) * wcap(c) <= 4
               fr * r1 * rcap(d) + fw * (w1 + w3) * wcap(d) <= 4
               fr * (1*r0 + 1*r1) + fw * (1*w0 + 1*w1 + 1*w2 + 1*w3) <= 5
        """
        # Create a variable for every read quorum and every write quorum. While
        # we do this, map each element x to the read and write quorums that
        # it's in. For example, image we have the following read and write
        # quorums:
        #
        #     read_quorums = [{a}, {a, b}, {a, c}]
        #     write_quorums = [{a, b}, {a, b, c}]
        #
        # Then, we'd have
        #
        #     read_quorum_vars = [r0, r1, 2]
        #     write_quorum_vars = [w0, w1]
        #     x_to_read_quorum_vars = {a: [r1, r2, r3], b: [r1], c: [r2]}
        #     x_to_write_quorum_vars = {a: [w1, w2], b: [w2, w2], c: [w2]}
        read_quorum_vars: List[pulp.LpVariable] = []
        x_to_read_quorum_vars: Dict[T, List[pulp.LpVariable]] = \
            collections.defaultdict(list)
        for (i, read_quorum) in enumerate(read_quorums):
            v = pulp.LpVariable(f'r{i}', 0, 1)
            read_quorum_vars.append(v)
            for x in read_quorum:
                x_to_read_quorum_vars[x].append(v)

        write_quorum_vars: List[pulp.LpVariable] = []
        x_to_write_quorum_vars: Dict[T, List[pulp.LpVariable]] = \
            collections.defaultdict(list)
        for (i, write_quorum) in enumerate(write_quorums):
            v = pulp.LpVariable(f'w{i}', 0, 1)
            write_quorum_vars.append(v)
            for x in write_quorum:
                x_to_write_quorum_vars[x].append(v)

        fr = sum(p * fr for (fr, p) in read_fraction.items())

        def network() -> pulp.LpAffineExpression:
            reads = fr * sum(
                v * len(rq)
                for (rq, v) in zip(read_quorums, read_quorum_vars)
            )
            writes = (1 - fr) * sum(
                v * len(wq)
                for (wq, v) in zip(write_quorums, write_quorum_vars)
            )
            return reads + writes

        def latency() -> pulp.LpAffineExpression:
            reads = fr * sum(
                v * self._read_quorum_latency(quorum).total_seconds()
                for (rq, v) in zip(read_quorums, read_quorum_vars)
                for quorum in [{self.node(x) for x in rq}]
            )
            writes = (1 - fr) * sum(
                v * self._write_quorum_latency(quorum).total_seconds()
                for (wq, v) in zip(write_quorums, write_quorum_vars)
                for quorum in [{self.node(x) for x in wq}]
            )
            return reads + writes

        def fr_load(problem: pulp.LpProblem, fr: float) -> pulp.LpAffineExpression:
            l = pulp.LpVariable(f'l_{fr}', 0, 1)

            for node in self.nodes():
                x = node.x
                x_load: pulp.LpAffineExpression = 0

                if x in x_to_read_quorum_vars:
                    vs = x_to_read_quorum_vars[x]
                    x_load += fr * sum(vs) / self.node(x).read_capacity

                if x in x_to_write_quorum_vars:
                    vs = x_to_write_quorum_vars[x]
                    x_load += (1 - fr) * sum(vs) / self.node(x).write_capacity

                problem += (x_load <= l, f'{x}{fr}')

            return l

        def load(problem: pulp.LpProblem,
                 read_fraction: Dict[float, float]) -> pulp.LpAffineExpression:
            return sum(p * fr_load(problem, fr)
                       for (fr, p) in read_fraction.items())

        # Form the linear program.
        problem = pulp.LpProblem("optimal_strategy", pulp.LpMinimize)

        # We add these constraints to make sure that the probabilities we
        # select form valid probabilty distributions.
        problem += (sum(read_quorum_vars) == 1, 'valid read strategy')
        problem += (sum(write_quorum_vars) == 1, 'valid write strategy')

        # Add the objective.
        if optimize == LOAD:
            problem += load(problem, read_fraction)
        elif optimize == NETWORK:
            problem += network()
        else:
            assert optimize == LATENCY
            problem += latency()

        # Add any constraints.
        if load_limit is not None:
            problem += (load(problem, read_fraction) <= load_limit,
                        'load limit')

        if network_limit is not None:
            problem += (network() <= network_limit, 'network limit')

        if latency_limit is not None:
            problem += (latency() <= latency_limit.total_seconds(),
                        'latency limit')

        # Solve the linear program.
        problem.solve(pulp.apis.PULP_CBC_CMD(msg=False))
        if problem.status != pulp.LpStatusOptimal:
            raise NoStrategyFoundError(
                'no strategy satisfies the given constraints')

        # Prune out any quorums with 0 probability.
        sigma_r = {
            frozenset(rq): v.varValue
            for (rq, v) in zip(read_quorums, read_quorum_vars)
            if v.varValue != 0
        }
        sigma_w = {
            frozenset(wq): v.varValue
            for (wq, v) in zip(write_quorums, write_quorum_vars)
            if v.varValue != 0
        }

        return Strategy(self, sigma_r, sigma_w)


class Strategy(Generic[T]):
    def __init__(self,
                 qs: QuorumSystem[T],
                 sigma_r: Dict[FrozenSet[T], float],
                 sigma_w: Dict[FrozenSet[T], float]) -> None:
        self.qs = qs
        self.sigma_r = sigma_r
        self.sigma_w = sigma_w

        # The probability that x is chosen as part of a read quorum.
        self.x_read_probability: Dict[T, float] = collections.defaultdict(float)
        for (read_quorum, p) in self.sigma_r.items():
            for x in read_quorum:
                self.x_read_probability[x] += p

        # The probability that x is chosen as part of a write quorum.
        self.x_write_probability: Dict[T, float] = collections.defaultdict(float)
        for (write_quorum, weight) in self.sigma_w.items():
            for x in write_quorum:
                self.x_write_probability[x] += weight

    @no_type_check
    def __str__(self) -> str:
        # T may not comparable, so mypy complains about this sort.
        reads = {tuple(sorted(rq)): p for (rq, p) in self.sigma_r.items()}
        writes = {tuple(sorted(wq)): p for (wq, p) in self.sigma_w.items()}
        return f'Strategy(reads={reads}, writes={writes})'

    def quorum_system(self) -> QuorumSystem[T]:
        return self.qs

    def node(self, x: T) -> Node[T]:
        return self.qs.node(x)

    def nodes(self) -> Set[Node[T]]:
        return self.qs.nodes()

    def get_read_quorum(self) -> Set[T]:
        return set(random.choices(list(self.sigma_r.keys()),
                                  weights=list(self.sigma_r.values()))[0])

    def get_write_quorum(self) -> Set[T]:
        return set(random.choices(list(self.sigma_w.keys()),
                                  weights=list(self.sigma_w.values()))[0])

    def load(self,
             read_fraction: Optional[Distribution] = None,
             write_fraction: Optional[Distribution] = None) -> float:
        d = distribution.canonicalize_rw(read_fraction, write_fraction)
        return sum(p * self._load(fr) for (fr, p) in d.items())

    def capacity(self,
                 read_fraction: Optional[Distribution] = None,
                 write_fraction: Optional[Distribution] = None) -> float:
        d = distribution.canonicalize_rw(read_fraction, write_fraction)
        return sum(p * 1 / self._load(fr) for (fr, p) in d.items())

    def network_load(self,
                     read_fraction: Optional[Distribution] = None,
                     write_fraction: Optional[Distribution] = None) -> float:
        d = distribution.canonicalize_rw(read_fraction, write_fraction)
        fr = sum(p * fr for (fr, p) in d.items())
        reads = fr * sum(p * len(rq) for (rq, p) in self.sigma_r.items())
        writes = (1 - fr) * sum(p * len(wq) for (wq, p) in self.sigma_w.items())
        return reads + writes

    # mypy doesn't like calling sum with timedeltas.
    @no_type_check
    def latency(self,
                read_fraction: Optional[Distribution] = None,
                write_fraction: Optional[Distribution] = None) \
                -> datetime.timedelta:
        d = distribution.canonicalize_rw(read_fraction, write_fraction)
        fr = sum(p * fr for (fr, p) in d.items())

        reads = fr * sum((
            p * self.qs._read_quorum_latency({self.node(x) for x in rq})
            for (rq, p) in self.sigma_r.items()
        ), datetime.timedelta(seconds=0))

        writes = (1 - fr) * sum((
            p * self.qs._write_quorum_latency({self.node(x) for x in wq})
            for (wq, p) in self.sigma_w.items()
        ), datetime.timedelta(seconds=0))

        return reads + writes

    def node_load(self,
                  node: Node[T],
                  read_fraction: Optional[Distribution] = None,
                  write_fraction: Optional[Distribution] = None) -> float:
        d = distribution.canonicalize_rw(read_fraction, write_fraction)
        return sum(p * self._node_load(node, fr) for (fr, p) in d.items())

    def node_utilization(self,
                         node: Node[T],
                         read_fraction: Optional[Distribution] = None,
                         write_fraction: Optional[Distribution] = None) \
                         -> float:
        d = distribution.canonicalize_rw(read_fraction, write_fraction)
        return sum(p * self._node_utilization(node, fr)
                   for (fr, p) in d.items())

    def node_throughput(self,
                        node: Node[T],
                        read_fraction: Optional[Distribution] = None,
                        write_fraction: Optional[Distribution] = None) -> float:
        d = distribution.canonicalize_rw(read_fraction, write_fraction)
        return sum(p * self._node_throughput(node, fr) for (fr, p) in d.items())

    def _load(self, fr: float) -> float:
        return max(self._node_load(node, fr) for node in self.nodes())

    def _node_load(self, node: Node[T], fr: float) -> float:
        fw = 1 - fr
        return (fr * self.x_read_probability[node.x] / node.read_capacity +
                fw * self.x_write_probability[node.x] / node.write_capacity)

    def _node_utilization(self, node: Node[T], fr: float) -> float:
        return self._node_load(node, fr) / self._load(fr)

    def _node_throughput(self, node: Node[T], fr: float) -> float:
        cap = 1 / self._load(fr)
        fw = 1 - fr
        return cap * (fr * self.x_read_probability[node.x] +
                      fw * self.x_write_probability[node.x])



from .distribution import Distribution
from .expr import choose, Expr, Node
from .quorum_system import (LATENCY, LOAD, NETWORK, NoStrategyFoundError,
                            QuorumSystem, Strategy, Tuple)
from typing import Iterator, List, Optional, TypeVar
import datetime
import itertools


T = TypeVar('T')


class NoQuorumSystemFoundError(ValueError):
    pass


def _partitionings(xs: List[T]) -> Iterator[List[List[T]]]:
    """
    _partitionings(xs) yields all possible partitionings of xs. For example,
    _partitionings([1, 2, 3]) yields

        [[1], [2], [3]]
        [[1, 2], [3]]
        [[1, 3], [2]]
        [[2, 3], [1]]
        [[1, 2, 3]]
    """
    if len(xs) == 0:
        return

    def helper(xs: List[T]) -> Iterator[List[List[T]]]:
        if len(xs) == 0:
            yield []
            return

        x = xs[0]
        for partition in helper(xs[1:]):
            yield [[x]] + partition
            for i in range(len(partition)):
                yield partition[:i] + [[x] + partition[i]] + partition[i+1:]

    yield from helper(xs)


def _dup_free_exprs(nodes: List[Node[T]],
                    max_height: int = 0) -> Iterator[Expr[T]]:
    """
    _dup_free_exprs(nodes) yields all possible duplicate free expressions over
    `nodes` with height at most `max_height`. If `max_height` is not positive,
    there is no height limit. Note that an expression might be yielded more
    than once.
    """
    assert len(nodes) > 0

    if len(nodes) == 1:
        yield nodes[0]
        return

    if max_height == 1:
        for k in range(1, len(nodes) + 1):
            yield choose(k, nodes) # type: ignore
        return

    for partitioning in _partitionings(nodes):
        # We ignore the partitioning that includes every node in a single
        # partition.
        if len(partitioning) == 1:
            continue

        subiterators = [_dup_free_exprs(p, max_height-1) for p in partitioning]
        for subexprs in itertools.product(*subiterators):
            for k in range(1, len(subexprs) + 1):
                yield choose(k, list(subexprs))


def search(nodes: List[Node[T]],
           read_fraction: Optional[Distribution] = None,
           write_fraction: Optional[Distribution] = None,
           optimize: str = LOAD,
           resilience: int = 0,
           load_limit: Optional[float] = None,
           network_limit: Optional[float] = None,
           latency_limit: Optional[datetime.timedelta] = None,
           f: int = 0,
           timeout: datetime.timedelta = datetime.timedelta(seconds=0)) \
           -> Tuple[QuorumSystem[T], Strategy[T]]:
    start_time = datetime.datetime.now()

    def metric(sigma: Strategy[T]) -> float:
        if optimize == LOAD:
            return sigma.load(read_fraction, write_fraction)
        elif optimize == NETWORK:
            return sigma.network_load(read_fraction, write_fraction)
        else:
            return sigma.latency(read_fraction, write_fraction).total_seconds()

    opt_qs: Optional[QuorumSystem[T]] = None
    opt_sigma: Optional[Strategy[T]] = None
    opt_metric: Optional[float] = None

    def do_search(exprs: Iterator[Expr[T]]) -> None:
        nonlocal opt_qs
        nonlocal opt_sigma
        nonlocal opt_metric

        for reads in exprs:
            qs = QuorumSystem(reads=reads)
            if qs.resilience() < resilience:
                continue

            try:
                sigma = qs.strategy(optimize = optimize,
                                    load_limit = load_limit,
                                    network_limit = network_limit,
                                    latency_limit = latency_limit,
                                    read_fraction = read_fraction,
                                    write_fraction = write_fraction,
                                    f = f)
                sigma_metric = metric(sigma)
                if opt_metric is None or sigma_metric < opt_metric:
                    opt_qs = qs
                    opt_sigma = sigma
                    opt_metric = sigma_metric
            except NoStrategyFoundError:
                pass

            if (timeout != datetime.timedelta(seconds=0) and
                datetime.datetime.now() - start_time >= timeout):
                return

    do_search(_dup_free_exprs(nodes, max_height=2))
    do_search(_dup_free_exprs(nodes))

    if opt_qs is None:
        raise ValueError('no quorum system found')
    else:
        assert opt_sigma is not None
        return (opt_qs, opt_sigma)



from quoracle import *
from quoracle.quorum_system import *
import unittest


class TestQuorumSystem(unittest.TestCase):
    def test_init(self):
        def quorums(e: Expr['str']) -> FrozenSet[FrozenSet[str]]:
            return frozenset(frozenset(q) for q in e.quorums())

        def assert_equal(x: Expr['str'], y: Expr['str']) -> None:
            self.assertEqual(quorums(x), quorums(y))

        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')

        # Specify reads.
        qs = QuorumSystem(reads = a + b)
        assert_equal(qs.reads, a + b)
        assert_equal(qs.writes, a * b)

        # Specify writes.
        qs = QuorumSystem(writes = a + b)
        assert_equal(qs.reads, a * b)
        assert_equal(qs.writes, a + b)

        # Specify neither.
        with self.assertRaises(ValueError):
            QuorumSystem()

        # Specify both overlapping.
        qs = QuorumSystem(reads=a+b, writes=a*b*c)
        assert_equal(qs.reads, a+b)
        assert_equal(qs.writes, a*b*c)

        # Specify both not overlapping.
        with self.assertRaises(ValueError):
            QuorumSystem(reads=a+b, writes=a)

    def test_uniform_strategy(self):
        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')

        sigma = QuorumSystem(reads=a).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a'}): 1.0,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a'}): 1.0,
        })

        sigma = QuorumSystem(reads=a+a).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a'}): 1.0,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a'}): 1.0,
        })

        sigma = QuorumSystem(reads=a*a).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a'}): 1.0,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a'}): 1.0,
        })

        sigma = QuorumSystem(reads=a + a*b).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a'}): 1.0,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a'}): 1.0,
        })

        sigma = QuorumSystem(reads=a + a*b + a*c).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a'}): 1.0,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a'}): 1.0,
        })

        sigma = QuorumSystem(reads=a + b).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a'}): 1 / 2,
            frozenset({'b'}): 1 / 2,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a', 'b'}): 1.0,
        })

        sigma = QuorumSystem(reads=a + b + c).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a'}): 1 / 3,
            frozenset({'b'}): 1 / 3,
            frozenset({'c'}): 1 / 3,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a', 'b', 'c'}): 1.0,
        })

        sigma = QuorumSystem(reads=(a*b)+(c*d)).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a', 'b'}): 1 / 2,
            frozenset({'c', 'd'}): 1 / 2,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a', 'c'}): 1 / 4,
            frozenset({'a', 'd'}): 1 / 4,
            frozenset({'b', 'c'}): 1 / 4,
            frozenset({'b', 'd'}): 1 / 4,
        })

        sigma = QuorumSystem(reads=(a*b)+(c*d)+(a*b)+(a*b*c)).uniform_strategy()
        self.assertEqual(sigma.sigma_r, {
            frozenset({'a', 'b'}): 1 / 2,
            frozenset({'c', 'd'}): 1 / 2,
        })
        self.assertEqual(sigma.sigma_w, {
            frozenset({'a', 'c'}): 1 / 4,
            frozenset({'a', 'd'}): 1 / 4,
            frozenset({'b', 'c'}): 1 / 4,
            frozenset({'b', 'd'}): 1 / 4,
        })

    def test_make_strategy(self):
        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')

        qs = QuorumSystem(reads = a*b + c*d)
        sigma = qs.make_strategy(
            sigma_r = {
                frozenset({'a', 'b'}): 25,
                frozenset({'c', 'd'}): 75,
            },
            sigma_w = {
                frozenset({'a', 'c'}): 1,
                frozenset({'a', 'd'}): 1,
                frozenset({'b', 'c'}): 1,
                frozenset({'b', 'd'}): 1,
            },
        )
        self.assertEqual(sigma.sigma_r,
            {
                frozenset({'a', 'b'}): 0.25,
                frozenset({'c', 'd'}): 0.75,
            },
        )
        self.assertEqual(sigma.sigma_w,
            {
                frozenset({'a', 'c'}): 0.25,
                frozenset({'a', 'd'}): 0.25,
                frozenset({'b', 'c'}): 0.25,
                frozenset({'b', 'd'}): 0.25,
            },
        )

        with self.assertRaises(ValueError):
            sigma = qs.make_strategy(
                sigma_r = {
                    frozenset({'a', 'b'}): -1,
                    frozenset({'c', 'd'}): 1,
                },
                sigma_w = {
                    frozenset({'a', 'c'}): 1,
                    frozenset({'a', 'd'}): 1,
                    frozenset({'b', 'c'}): 1,
                    frozenset({'b', 'd'}): 1,
                },
            )

        with self.assertRaises(ValueError):
            sigma = qs.make_strategy(
                sigma_r = {
                    frozenset({'a'}): 1,
                    frozenset({'c', 'd'}): 1,
                },
                sigma_w = {
                    frozenset({'a', 'c'}): 1,
                    frozenset({'a', 'd'}): 1,
                    frozenset({'b', 'c'}): 1,
                    frozenset({'b', 'd'}): 1,
                },
            )

    def test_optimal_strategy(self):
        def s(n: int) -> datetime.timedelta:
            return datetime.timedelta(seconds=n)

        a = Node('a', write_capacity=1, read_capacity=2, latency=s(1))
        b = Node('b', write_capacity=1, read_capacity=2, latency=s(2))
        c = Node('c', write_capacity=1, read_capacity=2, latency=s(3))
        d = Node('d', write_capacity=1, read_capacity=2, latency=s(4))
        qs = QuorumSystem(reads=a*b + c*d)

        # Load Optimized.
        self.assertEqual(qs.load(read_fraction=1), 0.25)
        self.assertEqual(qs.capacity(read_fraction=1), 4)
        self.assertEqual(qs.load(read_fraction=0), 0.5)
        self.assertEqual(qs.capacity(read_fraction=0), 2)

        self.assertEqual(qs.load(read_fraction=1, network_limit=2), 0.25)
        self.assertEqual(qs.capacity(read_fraction=1, network_limit=2), 4)
        self.assertEqual(qs.load(read_fraction=0, network_limit=2), 0.5)
        self.assertEqual(qs.capacity(read_fraction=0, network_limit=2), 2)

        self.assertEqual(qs.load(read_fraction=1, latency_limit=s(4)), 0.25)
        self.assertEqual(qs.capacity(read_fraction=1, latency_limit=s(4)), 4)
        self.assertEqual(qs.load(read_fraction=0, latency_limit=s(4)), 0.5)
        self.assertEqual(qs.capacity(read_fraction=0, latency_limit=s(4)), 2)

        # Network Optimized.
        self.assertEqual(qs.network_load(
            read_fraction=1,
            optimize='network',
        ), 2)
        self.assertEqual(qs.network_load(
            read_fraction=0,
            optimize='network',
        ), 2)
        self.assertEqual(qs.network_load(
            read_fraction=1,
            optimize='network',
            load_limit = 0.25,
        ), 2)
        self.assertEqual(qs.network_load(
            read_fraction=0,
            optimize='network',
            load_limit = 0.5,
        ), 2)
        self.assertEqual(qs.network_load(
            read_fraction=1,
            optimize='network',
            latency_limit = s(2),
        ), 2)
        self.assertEqual(qs.network_load(
            read_fraction=0,
            optimize='network',
            latency_limit = s(3),
        ), 2)

        # Latency Optimized.
        self.assertEqual(qs.latency(read_fraction=1, optimize='latency'), s(2))
        self.assertEqual(qs.latency(read_fraction=0, optimize='latency'), s(3))
        self.assertEqual(qs.latency(
            read_fraction=1,
            optimize='latency',
            load_limit = 1.0,
        ), s(2))
        self.assertEqual(qs.latency(
            read_fraction=0,
            optimize='latency',
            load_limit = 1.0,
        ), s(3))
        self.assertEqual(qs.latency(
            read_fraction=1,
            optimize='latency',
            network_limit = 2,
        ), s(2))
        self.assertEqual(qs.latency(
            read_fraction=0,
            optimize='latency',
            network_limit = 2,
        ), s(3))

        # 1-Resilient Load Optimized.
        self.assertEqual(qs.load(read_fraction=1, f=1), 0.5)
        self.assertEqual(qs.capacity(read_fraction=1, f=1), 2)
        self.assertEqual(qs.load(read_fraction=0, f=1), 1)
        self.assertEqual(qs.capacity(read_fraction=0, f=1), 1)

        # 1-Resilient Network Optimized.
        self.assertEqual(
            qs.network_load(read_fraction=1, optimize='network', f=1), 4)
        self.assertEqual(
            qs.network_load(read_fraction=0, optimize='network', f=1), 4)

        # 1-Resilient Latency Optimized.
        self.assertEqual(
            qs.latency(read_fraction=1, optimize='latency', f=1), s(2))
        self.assertEqual(
            qs.latency(read_fraction=0, optimize='latency', f=1), s(3))

        # Illegal Specification
        with self.assertRaises(ValueError):
            qs.strategy(read_fraction=0.1, optimize='load', load_limit=1)

        with self.assertRaises(ValueError):
            qs.strategy(read_fraction=0.1, optimize='network', network_limit=2)

        with self.assertRaises(ValueError):
            qs.strategy(read_fraction=0.1, optimize='latency',
                        latency_limit=s(5))

        # Unsatisfiable Constraints
        with self.assertRaises(NoStrategyFoundError):
            qs.strategy(read_fraction=0,
                        optimize='load',
                        network_limit=1.5)

        with self.assertRaises(NoStrategyFoundError):
            qs.strategy(read_fraction=0,
                        optimize='load',
                        latency_limit=s(1))

        with self.assertRaises(NoStrategyFoundError):
            qs.strategy(read_fraction=1,
                        optimize='network',
                        load_limit=0.25,
                        latency_limit=s(2))



from typing import Any, Callable, List, NamedTuple, Optional, Tuple
import math


class Point(NamedTuple):
    x: float
    y: float


class Segment:
    def __init__(self, l: Point, r: Point) -> None:
        assert l != r
        assert l.x < r.x
        self.l = l
        self.r = r

    def __str__(self) -> str:
        return f'{tuple(self.l)} -> {tuple(self.r)}'

    def __repr__(self) -> str:
        return f'Segment({self.l}, {self.r})'

    def __eq__(self, other) -> bool:
        if isinstance(other, Segment):
            return (self.l, self.r) == (other.l, other.r)
        else:
            return False

    def __hash__(self) -> int:
        return hash((self.l, self.r))

    def __call__(self, x: float) -> float:
        assert self.l.x <= x <= self.r.x
        return self.slope() * (x - self.l.x) + self.l.y

    def approximately_equal(self, other: 'Segment') -> float:
        return (math.isclose(self.l.y, other.l.y, rel_tol=1e-5) and
                math.isclose(self.r.y, other.r.y, rel_tol=1e-5))

    def compatible(self, other: 'Segment') -> float:
        return self.l.x == other.l.x and self.r.x == other.r.x

    def slope(self) -> float:
        return (self.r.y - self.l.y) / (self.r.x - self.l.x)

    def above(self, other: 'Segment') -> bool:
        assert self.compatible(other)
        return self != other and self.l.y >= other.l.y and self.r.y >= other.r.y

    def above_eq(self, other: 'Segment') -> bool:
        assert self.compatible(other)
        return self == other or self.above(other)

    def intersects(self, other: 'Segment') -> bool:
        assert self.compatible(other)

        if self == other:
            return True
        elif self.l.y == other.l.y or self.r.y == other.r.y:
            return True
        elif self.above(other) or other.above(self):
            return False
        else:
            return True

    def intersection(self, other: 'Segment') -> Optional[Point]:
        assert self.compatible(other)

        if self == other or not self.intersects(other):
            return None

        x = ((other.l.y - self.l.y) /
             (self.r.y - other.r.y + other.l.y - self.l.y))
        return Point(x, self(x))


def max_of_segments(segments: List[Segment]) -> List[Tuple[float, float]]:
    assert len(segments) > 0
    assert len({segment.l.x for segment in segments}) == 1
    assert len({segment.r.x for segment in segments}) == 1

    # We compute the x-coordinate of every intersection point. We sort the
    # x-coordinates and for every x, we compute the highest line at that point.
    xs: List[float] = [0.0, 1.0]
    for (i, s1) in enumerate(segments):
        for (j, s2) in enumerate(segments[i + 1:], i + 1):
            p = s1.intersection(s2)
            if p is not None:
                xs.append(p.x)
    xs.sort()
    return [(x, max(segments, key=lambda s: s(x))(x)) for x in xs]



from quoracle import *
from quoracle.expr import *
from typing import Any, FrozenSet
import unittest

class TestExpr(unittest.TestCase):
    def test_quorums(self):
        def assert_equal(e: Expr[str], xs: List[Set[str]]) -> None:
            self.assertEqual(frozenset(frozenset(q) for q in e.quorums()),
                             frozenset(frozenset(x) for x in xs))
        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')
        e = Node('e')
        f = Node('f')

        assert_equal(a+b+c, [{'a'}, {'b'}, {'c'}])
        assert_equal(a*b*c, [{'a', 'b', 'c'}])
        assert_equal(a + b*c, [{'a'}, {'b', 'c'}])
        assert_equal(a*a*a, [{'a'}])
        assert_equal(a+a+a, [{'a'}])
        assert_equal(a*(a+b), [{'a'}, {'a', 'b'}])
        assert_equal(choose(1, [a, b, c]), [{'a'}, {'b'}, {'c'}])
        assert_equal(choose(2, [a, b, c]), [{'a', 'b'}, {'a', 'c'}, {'b', 'c'}])
        assert_equal(choose(3, [a, b, c]), [{'a', 'b', 'c'}])
        assert_equal((a+b) * (c+d), [
            {'a', 'c'}, {'a', 'd'}, {'b', 'c'}, {'b', 'd'}
        ])
        assert_equal((a+b) * (a+c), [
            {'a'}, {'a', 'c'}, {'a', 'b'}, {'b', 'c'}
        ])
        assert_equal(
            choose(2, [
                choose(2, [a, b, c]),
                choose(2, [d, e, f]),
                choose(2, [a, c, e]),
            ]), [
                {'a','b','d','e'}, {'a','b','d','f'}, {'a','b','e','f'},
                {'a','c','d','e'}, {'a','c','d','f'}, {'a','c','e','f'},
                {'b','c','d','e'}, {'b','c','d','f'}, {'b','c','e','f'},
                {'a','b','a','c'}, {'a','b','a','e'}, {'a','b','c','e'},
                {'a','c','a','c'}, {'a','c','a','e'}, {'a','c','c','e'},
                {'b','c','a','c'}, {'b','c','a','e'}, {'b','c','c','e'},
                {'d','e','a','c'}, {'d','e','a','e'}, {'d','e','c','e'},
                {'d','f','a','c'}, {'d','f','a','e'}, {'d','f','c','e'},
                {'e','f','a','c'}, {'e','f','a','e'}, {'e','f','c','e'},
        ])

    def test_is_quorum(self):
        def assert_quorum(e: Expr[str], q: Set[str]) -> None:
            self.assertTrue(e.is_quorum(q))

        def assert_not_quorum(e: Expr[str], q: Set[str]) -> None:
            self.assertFalse(e.is_quorum(q))

        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')
        e = Node('e')
        f = Node('f')

        expr = a + b + c
        assert_quorum(expr, {'a'})
        assert_quorum(expr, {'b'})
        assert_quorum(expr, {'c'})
        assert_quorum(expr, {'a', 'b'})
        assert_quorum(expr, {'a', 'c'})
        assert_quorum(expr, {'b', 'c'})
        assert_quorum(expr, {'a', 'b', 'c'})
        assert_quorum(expr, {'a', 'x'})
        assert_not_quorum(expr, set())
        assert_not_quorum(expr, {'x'})

        expr = a * b * c
        assert_quorum(expr, {'a', 'b', 'c'})
        assert_quorum(expr, {'a', 'b', 'c', 'x'})
        assert_not_quorum(expr, set())
        assert_not_quorum(expr, {'a'})
        assert_not_quorum(expr, {'b'})
        assert_not_quorum(expr, {'c'})
        assert_not_quorum(expr, {'a', 'b'})
        assert_not_quorum(expr, {'a', 'c'})
        assert_not_quorum(expr, {'b', 'c'})
        assert_not_quorum(expr, {'x'})
        assert_not_quorum(expr, {'a', 'x'})

        expr = choose(2, [a, b, c])
        assert_quorum(expr, {'a', 'b'})
        assert_quorum(expr, {'a', 'c'})
        assert_quorum(expr, {'b', 'c'})
        assert_quorum(expr, {'a', 'b', 'c'})
        assert_quorum(expr, {'a', 'b', 'c', 'x'})
        assert_not_quorum(expr, {'a'})
        assert_not_quorum(expr, {'b'})
        assert_not_quorum(expr, {'c'})
        assert_not_quorum(expr, {'x'})

        expr = (a+b) * (c+d)
        assert_quorum(expr, {'a', 'c'})
        assert_quorum(expr, {'a', 'c'})
        assert_quorum(expr, {'a', 'd'})
        assert_quorum(expr, {'b', 'd'})
        assert_quorum(expr, {'a', 'b', 'd'})
        assert_quorum(expr, {'b', 'c', 'd'})
        assert_quorum(expr, {'a', 'b', 'd'})
        assert_quorum(expr, {'a', 'c', 'd'})
        assert_quorum(expr, {'a', 'b', 'd'})
        assert_quorum(expr, {'b', 'c', 'd'})
        assert_quorum(expr, {'a', 'b', 'd'})
        assert_quorum(expr, {'b', 'c', 'd'})
        assert_quorum(expr, {'a', 'b', 'c', 'd'})
        assert_not_quorum(expr, {'a'})
        assert_not_quorum(expr, {'b'})
        assert_not_quorum(expr, {'c'})
        assert_not_quorum(expr, {'d'})
        assert_not_quorum(expr, {'a', 'b'})
        assert_not_quorum(expr, {'c', 'd'})
        assert_not_quorum(expr, {'a', 'b', 'x'})

    def test_resilience(self):
        def assert_resilience(e: Expr[str], n: int):
            self.assertEqual(e.resilience(), n)

        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')
        e = Node('e')
        f = Node('f')

        assert_resilience(a, 0)
        assert_resilience(a + b, 1)
        assert_resilience(a + b + c, 2)
        assert_resilience(a + b + c + d, 3)
        assert_resilience(a, 0)
        assert_resilience(a * b, 0)
        assert_resilience(a * b * c, 0)
        assert_resilience(a * b * c * d, 0)
        assert_resilience((a + b) * (c + d), 1)
        assert_resilience((a + b + c) * (d + e + f), 2)
        assert_resilience((a + b + c) * (a + e + f), 2)
        assert_resilience((a + a + c) * (d + e + f), 1)
        assert_resilience((a + a + a) * (d + e + f), 0)
        assert_resilience(a*b + b*c + a*d + a*d*e, 1)
        assert_resilience(choose(2, [a, b, c]), 1)
        assert_resilience(choose(2, [a, b, c, d, e]), 3)
        assert_resilience(choose(3, [a, b, c, d, e]), 2)
        assert_resilience(choose(4, [a, b, c, d, e]), 1)
        assert_resilience(choose(2, [a+b+c, d+e, f]), 2)
        assert_resilience(choose(2, [a*b, a*c, d]), 0)
        assert_resilience(choose(2, [a+b, a+c, a+d]), 2)

    def test_dual(self):
        def quorums(e: Expr['str']) -> FrozenSet[FrozenSet[str]]:
            return frozenset(frozenset(q) for q in e.quorums())

        def assert_equal(x: Expr['str'], y: Expr['str']) -> None:
            self.assertEqual(quorums(x), quorums(y))

        def assert_dual(x: Expr['str'], y: Expr['str']) -> None:
            assert_equal(x.dual(), y)

        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')
        e = Node('e')

        assert_dual(a, a)
        assert_dual(a + b, a * b)
        assert_dual(a + a, a * a)
        assert_dual((a + b) * (c + d), (a * b) + (c * d))
        assert_dual((a + b) * (a + d), (a * b) + (a * d))
        assert_dual((a + b) * (a + a), (a * b) + (a * a))
        assert_dual((a + a) * (a + a), (a * a) + (a * a))
        assert_dual((a + (a * b)) + ((c * d) + a),
                    (a * (a + b)) * ((c + d) * a))
        assert_dual(choose(2, [a, b, c]), choose(2, [a, b, c]))
        assert_dual(choose(2, [a+b, c+d, e]), choose(2, [a*b, c*d, e]))
        assert_dual(choose(3, [a, b, c, d, e]), choose(3, [a, b, c, d, e]))
        assert_dual(choose(2, [a, b, c, d, e]), choose(4, [a, b, c, d, e]))
        assert_dual(choose(4, [a, b, c, d, e]), choose(2, [a, b, c, d, e]))

    def test_dup_free(self):
        def assert_dup_free(e: Expr['str']) -> None:
            self.assertTrue(e.dup_free())

        def assert_not_dup_free(e: Expr['str']) -> None:
            self.assertFalse(e.dup_free())

        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')
        e = Node('e')
        f = Node('f')

        assert_dup_free(a)
        assert_dup_free(a + b)
        assert_dup_free(a * b)
        assert_dup_free(a * b + c)
        assert_dup_free(choose(2, [a, b, c]))
        assert_dup_free(choose(2, [a*b, c, d+e+f]))
        assert_dup_free(choose(3, [a, b, c, d, e]))
        assert_dup_free((a + b) * (c + (d * e)))

        assert_not_dup_free(a + a)
        assert_not_dup_free(a * a)
        assert_not_dup_free(a * (b + a))
        assert_not_dup_free(choose(2, [a, b, a]))
        assert_not_dup_free(choose(3, [a, b, c, d, a]))
        assert_not_dup_free((a + b) * (c + (d * a)))



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="quoracle",
    version="0.0.3",
    author="Michael Whittaker",
    author_email="mwhittttaker@gmail.com",
    description=("A library for modelling, analyzing, and optimizing quorum " +
                 "systems"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mwhittaker/quorums",
    packages=setuptools.find_packages(),
    install_requires =[
        "matplotlib",
        "pulp",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from quoracle import *
from quoracle.expr import Expr
from quoracle.search import _dup_free_exprs, _partitionings
from typing import Any, FrozenSet, List
import datetime
import unittest


class TestSearch(unittest.TestCase):
    def test_partitions(self):
        def setify(partitions: List[List[List[Any]]]) \
                   -> FrozenSet[FrozenSet[FrozenSet[Any]]]:
            return frozenset(frozenset(frozenset(s) for s in partition)
                             for partition in partitions)

        def assert_equal(x: List[List[Any]], y: List[List[Any]]):
            self.assertEqual(setify(x), setify(y))

        assert_equal(list(_partitionings([])), [])
        assert_equal(list(_partitionings([1])), [[[1]]])
        assert_equal(list(_partitionings([1, 2])), [
            [[1], [2]],
            [[1, 2]],
        ])
        assert_equal(list(_partitionings([1, 2, 3])), [
            [[1], [2], [3]],
            [[1, 2], [3]],
            [[1, 3], [2]],
            [[2, 3], [1]],
            [[1, 2, 3]],
        ])
        assert_equal(list(_partitionings([1, 2, 3, 4])), [
            [[1], [2], [3], [4]],
            [[1, 2], [3], [4]],
            [[1, 3], [2], [4]],
            [[1, 4], [2], [3]],
            [[2, 3], [1], [4]],
            [[2, 4], [1], [3]],
            [[3, 4], [1], [2]],
            [[1, 2], [3, 4]],
            [[1, 3], [2, 4]],
            [[1, 4], [2, 3]],
            [[2, 3, 4], [1]],
            [[1, 3, 4], [2]],
            [[1, 2, 4], [3]],
            [[1, 2, 3], [4]],
            [[1, 2, 3, 4]],
        ])

    def test_dup_free_exprs(self):
        def quorums(e: Expr) -> FrozenSet[FrozenSet[Any]]:
            return frozenset(frozenset(q) for q in e.quorums())

        def assert_equal(xs: List[Expr], ys: List[Expr]) -> None:
            self.assertEqual(frozenset(quorums(x) for x in xs),
                             frozenset(quorums(y) for y in ys))

        a = Node('a')
        b = Node('b')
        c = Node('c')
        d = Node('d')

        assert_equal(list(_dup_free_exprs([a])), [a])
        assert_equal(list(_dup_free_exprs([a, b])), [
            a + b,
            a * b,
        ])
        assert_equal(list(_dup_free_exprs([a, b, c])), [
            a + b + c,
            choose(2, [a, b, c]),
            a * b * c,
            (a + b) + c,
            (a + b) * c,
            (a * b) + c,
            (a * b) * c,
            (a + c) + b,
            (a + c) * b,
            (a * c) + b,
            (a * c) * b,
            (b + c) + a,
            (b + c) * a,
            (b * c) + a,
            (b * c) * a,
        ])
        assert_equal(list(_dup_free_exprs([a, b, c], max_height=1)), [
            a + b + c,
            choose(2, [a, b, c]),
            a * b * c,
        ])
        assert_equal(list(_dup_free_exprs([a, b, c, d], max_height=1)), [
            a + b + c + d,
            choose(2, [a, b, c, d]),
            choose(3, [a, b, c, d]),
            a * b * c * d,
        ])
        assert_equal(list(_dup_free_exprs([a, b, c, d], max_height=2)), [
            a + b + c + d,
            choose(2, [a, b, c, d]),
            choose(3, [a, b, c, d]),
            a * b * c * d,

            (a + b) + c + d,
            (a + b) * c * d,
            (a * b) + c + d,
            (a * b) * c * d,
            choose(2, [a + b, c, d]),
            choose(2, [a * b, c, d]),

            (a + c) + b + d,
            (a + c) * b * d,
            (a * c) + b + d,
            (a * c) * b * d,
            choose(2, [a + c, b, d]),
            choose(2, [a * c, b, d]),

            (a + d) + b + c,
            (a + d) * b * c,
            (a * d) + b + c,
            (a * d) * b * c,
            choose(2, [a + d, b, c]),
            choose(2, [a * d, b, c]),

            (b + c) + a + d,
            (b + c) * a * d,
            (b * c) + a + d,
            (b * c) * a * d,
            choose(2, [b + c, a, d]),
            choose(2, [b * c, a, d]),

            (b + d) + a + c,
            (b + d) * a * c,
            (b * d) + a + c,
            (b * d) * a * c,
            choose(2, [b + d, a, c]),
            choose(2, [b * d, a, c]),

            (c + d) + a + b,
            (c + d) * a * b,
            (c * d) + a + b,
            (c * d) * a * b,
            choose(2, [c + d, a, b]),
            choose(2, [c * d, a, b]),

            (a + b) + (c + d),
            (a + b) + (c * d),
            (a + b) * (c + d),
            (a + b) * (c * d),
            (a * b) + (c + d),
            (a * b) + (c * d),
            (a * b) * (c + d),
            (a * b) * (c * d),

            (a + c) + (b + d),
            (a + c) + (b * d),
            (a + c) * (b + d),
            (a + c) * (b * d),
            (a * c) + (b + d),
            (a * c) + (b * d),
            (a * c) * (b + d),
            (a * c) * (b * d),

            (a + d) + (b + c),
            (a + d) + (b * c),
            (a + d) * (b + c),
            (a + d) * (b * c),
            (a * d) + (b + c),
            (a * d) + (b * c),
            (a * d) * (b + c),
            (a * d) * (b * c),

            a + (b + c + d),
            a + (b * c * d),
            a * (b + c + d),
            a * (b * c * d),
            a + choose(2, [b, c, d]),
            a * choose(2, [b, c, d]),

            b + (a + c + d),
            b + (a * c * d),
            b * (a + c + d),
            b * (a * c * d),
            b + choose(2, [a, c, d]),
            b * choose(2, [a, c, d]),

            c + (a + b + d),
            c + (a * b * d),
            c * (a + b + d),
            c * (a * b * d),
            c + choose(2, [a, b, d]),
            c * choose(2, [a, b, d]),

            d + (a + b + c),
            d + (a * b * c),
            d * (a + b + c),
            d * (a * b * c),
            d + choose(2, [a, b, c]),
            d * choose(2, [a, b, c]),
        ])

    def test_search(self):
        a = Node('a', capacity=1, latency=datetime.timedelta(seconds=2))
        b = Node('b', capacity=2, latency=datetime.timedelta(seconds=1))
        c = Node('c', capacity=1, latency=datetime.timedelta(seconds=2))
        d = Node('d', capacity=2, latency=datetime.timedelta(seconds=1))
        e = Node('e', capacity=1, latency=datetime.timedelta(seconds=2))
        f = Node('f', capacity=2, latency=datetime.timedelta(seconds=1))

        for fr in [0, 0.5, 1]:
            search([a, b, c], read_fraction=fr)
            search([a, b, c], read_fraction=fr, optimize='network')
            search([a, b, c], read_fraction=fr, optimize='latency')
            search([a, b, c], read_fraction=fr, resilience=1)
            search([a, b, c], read_fraction=fr, f=1)

        search([a, b, c],
               read_fraction=0.25,
               network_limit=3,
               latency_limit=datetime.timedelta(seconds=2))

        for fr in [0, 0.5]:
            t = datetime.timedelta(seconds=0.25)
            nodes = [a, b, c, d, e, f]
            search(nodes, read_fraction=fr, timeout=t)
            search(nodes, read_fraction=fr, timeout=t, optimize='network')
            search(nodes, read_fraction=fr, timeout=t, optimize='latency')
            search(nodes, read_fraction=fr, timeout=t, resilience=1)
            search(nodes, read_fraction=fr, timeout=t, f=1)



from typing import Dict, Iterator, Generic, List, Optional, Set, TypeVar
import datetime
import itertools
import pulp


T = TypeVar('T')


def _min_hitting_set(sets: Iterator[Set[T]]) -> int:
    x_vars: Dict[T, pulp.LpVariable] = dict()
    next_id = itertools.count()

    problem = pulp.LpProblem("min_hitting_set", pulp.LpMinimize)
    for (i, xs) in enumerate(sets):
        for x in xs:
            if x not in x_vars:
                id = next(next_id)
                x_vars[x] = pulp.LpVariable(f'x{id}', cat=pulp.LpBinary)
        problem += sum(x_vars[x] for x in xs) >= 1

    problem += sum(x_vars.values())
    problem.solve(pulp.apis.PULP_CBC_CMD(msg=False))
    return int(sum(v.varValue for v in x_vars.values()))


class Expr(Generic[T]):
    def __add__(self, rhs: 'Expr[T]') -> 'Expr[T]':
        def _or(lhs: Expr[T], rhs: Expr[T]) -> 'Or[T]':
            if isinstance(lhs, Or) and isinstance(rhs, Or):
                return Or(lhs.es + rhs.es)
            elif isinstance(lhs, Or):
                return Or(lhs.es + [rhs])
            elif isinstance(rhs, Or):
                return Or([lhs] + rhs.es)
            else:
                return Or([lhs, rhs])


        return _or(self, rhs)

    def __mul__(self, rhs: 'Expr[T]') -> 'Expr[T]':
        def _and(lhs: Expr[T], rhs: Expr[T]) -> 'And[T]':
            if isinstance(lhs, And) and isinstance(rhs, And):
                return And(lhs.es + rhs.es)
            elif isinstance(lhs, And):
                return And(lhs.es + [rhs])
            elif isinstance(rhs, And):
                return And([lhs] + rhs.es)
            else:
                return And([lhs, rhs])

        return _and(self, rhs)

    def quorums(self) -> Iterator[Set[T]]:
        raise NotImplementedError

    # TODO(mwhittaker): Add a function to return minimal quorums.

    # TODO(mwhittaker): Add a function to check whether two expressions are
    # equal. One simple way to do this is compare the set of minimal quorums.
    # There might be more efficient ways to check if two expressions are equal.

    def is_quorum(self, xs: Set[T]) -> bool:
        raise NotImplementedError

    def elements(self) -> Set[T]:
        return {node.x for node in self.nodes()}

    def nodes(self) -> Set['Node[T]']:
        raise NotImplementedError

    def resilience(self) -> int:
        if self.dup_free():
            return self._dup_free_min_failures() - 1
        else:
            return _min_hitting_set(self.quorums()) - 1

    def dual(self) -> 'Expr[T]':
        raise NotImplementedError

    def dup_free(self) -> bool:
        return len(self.nodes()) == self._num_leaves()

    def _num_leaves(self) -> int:
        raise NotImplementedError

    def _dup_free_min_failures(self) -> int:
        raise NotImplementedError


class Node(Expr[T]):
    def __init__(self,
                 x: T,
                 capacity: Optional[float] = None,
                 read_capacity: Optional[float] = None,
                 write_capacity: Optional[float] = None,
                 latency: datetime.timedelta = None) -> None:
        self.x = x

        # A user either specifies capacity or (read_capacity and
        # write_capacity), but not both.
        if (capacity is None and
            read_capacity is None and
            write_capacity is None):
            self.read_capacity = 1.0
            self.write_capacity = 1.0
        elif (capacity is not None and
              read_capacity is None and
              write_capacity is None):
            self.read_capacity = capacity
            self.write_capacity = capacity
        elif (capacity is None and
              read_capacity is not None and
              write_capacity is not None):
            self.read_capacity = read_capacity
            self.write_capacity = write_capacity
        else:
            raise ValueError('You must specify capacity or (read_capacity '
                             'and write_capacity)')

        if latency is None:
            self.latency = datetime.timedelta(seconds=1)
        else:
            self.latency = latency


    def __str__(self) -> str:
        return str(self.x)

    def __repr__(self) -> str:
        return f'Node({self.x})'

    def quorums(self) -> Iterator[Set[T]]:
        yield {self.x}

    def is_quorum(self, xs: Set[T]) -> bool:
        return self.x in xs

    def nodes(self) -> Set['Node[T]']:
        return {self}

    def dual(self) -> Expr:
        return self

    def _num_leaves(self) -> int:
        return 1

    def _dup_free_min_failures(self) -> int:
        return 1


class Or(Expr[T]):
    def __init__(self, es: List[Expr[T]]) -> None:
        if len(es) == 0:
            raise ValueError(f'Or cannot be constructed with an empty list')

        self.es = es

    def __str__(self) -> str:
        return '(' + ' + '.join(str(e) for e in self.es) + ')'

    def __repr__(self) -> str:
        return f'Or({self.es})'

    def quorums(self) -> Iterator[Set[T]]:
        for e in self.es:
            yield from e.quorums()

    def is_quorum(self, xs: Set[T]) -> bool:
        return any(e.is_quorum(xs) for e in self.es)

    def nodes(self) -> Set[Node[T]]:
        return set.union(*[e.nodes() for e in self.es])

    def dual(self) -> Expr:
        return And([e.dual() for e in self.es])

    def _num_leaves(self) -> int:
        return sum(e._num_leaves() for e in self.es)

    def _dup_free_min_failures(self) -> int:
        return sum(e._dup_free_min_failures() for e in self.es)


class And(Expr[T]):
    def __init__(self, es: List[Expr[T]]) -> None:
        if len(es) == 0:
            raise ValueError(f'And cannot be constructed with an empty list')

        self.es = es

    def __str__(self) -> str:
        return '(' + ' * '.join(str(e) for e in self.es) + ')'

    def __repr__(self) -> str:
        return f'And({self.es})'

    def quorums(self) -> Iterator[Set[T]]:
        for subquorums in itertools.product(*[e.quorums() for e in self.es]):
            yield set.union(*subquorums)

    def is_quorum(self, xs: Set[T]) -> bool:
        return all(e.is_quorum(xs) for e in self.es)

    def nodes(self) -> Set[Node[T]]:
        return set.union(*[e.nodes() for e in self.es])

    def dual(self) -> Expr:
        return Or([e.dual() for e in self.es])

    def _num_leaves(self) -> int:
        return sum(e._num_leaves() for e in self.es)

    def _dup_free_min_failures(self) -> int:
        return min(e._dup_free_min_failures() for e in self.es)

class Choose(Expr[T]):
    def __init__(self, k: int, es: List[Expr[T]]) -> None:
        if k <= 0 or k > len(es):
            raise ValueError(f'k must be in the range [1, {len(es)}]')

        self.k = k
        self.es = es

    def __str__(self) -> str:
        return f'choose{self.k}(' + ', '.join(str(e) for e in self.es) + ')'

    def __repr__(self) -> str:
        return f'Chose({self.k}, {self.es})'

    def quorums(self) -> Iterator[Set[T]]:
        for combo in itertools.combinations(self.es, self.k):
            for subquorums in itertools.product(*[e.quorums() for e in combo]):
                yield set.union(*subquorums)

    def is_quorum(self, xs: Set[T]) -> bool:
        return sum(1 if e.is_quorum(xs) else 0 for e in self.es) >= self.k

    def nodes(self) -> Set[Node[T]]:
        return set.union(*[e.nodes() for e in self.es])

    def dual(self) -> Expr:
        # TODO(mwhittaker): Prove that this is in fact the dual.
        return Choose(len(self.es) - self.k + 1, [e.dual() for e in self.es])

    def _num_leaves(self) -> int:
        return sum(e._num_leaves() for e in self.es)

    def _dup_free_min_failures(self) -> int:
        subfailures = [e._dup_free_min_failures() for e in self.es]
        return sum(sorted(subfailures)[:len(subfailures) - self.k + 1])


def choose(k: int, es: List[Expr[T]]) -> Expr[T]:
    if len(es) == 0:
        raise ValueError('no expressions provided')

    if not (1 <= k <= len(es)):
        raise ValueError('k must be in the range [1, len(es)]')

    if k == 1:
        return Or(es)
    elif k == len(es):
        return And(es)
    else:
        return Choose(k, es)


def majority(es: List[Expr[T]]) -> Expr[T]:
    if len(es) == 0:
        raise ValueError('no expressions provided')

    return choose(len(es) // 2 + 1, es)
